/*
 * $Id: midilib.h,v 1.41 1998/06/04 05:33:26 keithp Exp $
 *
 * Modified by W.H.Scholten (dd 2000 - 2001)
 *
 * Copyright  1998 Keith Packard
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Keith Packard not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Keith Packard makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef _GSI__gsimidi_pcm__kpmidi_h
#define _GSI__gsimidi_pcm__kpmidi_h


#include <gsi/whs_types.h>
#include <gsi/kpmidi_patch.h>

typedef int  Boolean;

typedef uint32  MidiSine;

/* also in gsimidi_pcm/midi_data.h: */
#define MIN_DRUM    35
#define MAX_DRUM    81


//NOT USED enum { NOTE_OFF = 8, NOTE_ON, AFTER_TOUCH, CONTROL_CHANGE, PROGRAM_CHANGE,
//NOT USED    CHANNEL_PRESSURE, PITCH_WHEEL, SYSEX };

typedef struct _voiceWave {
    MidiFix	    sample;
    MidiFix	    sampleStep;	/* wave sample step per output sample */
    MidiFix	    end;	/* something happens here */
    MidiFix	    loop_len;	/* loop is this long */
    MidiFix	    data_end;	/* end of sample */
    Boolean	    loopRelease;/* use loop during release */
    double	    freqFactor;	/* sample rate conversion factor */
    const float	    *samples;	/* samples */
} VoiceWave;


#define MIDI_CHANNELS	16

typedef struct _voiceEnvelope {
    float	    fLeft;
    float	    fRight;
    float	    fValue;
    float	    fPan;
    float	    fTrem;
    int		    remain;	/* samples until next segment */
    int32	    valueStep;	/* envel value step per output sample */
    int32	    value;	/* current envelope value */
    int32	    target;	/* goal value for envelope element */
    Boolean	    on;		/* voice sounding */
    int		    element;	/* index into array of envelope values */
    double	    envelopeRatio; /* ENVELOPE_RATE / output envelopeRate */
    float       volume; /* velocity * channel volume in range 0-1 */
    float	    in_scale;	/* instrument scale factor */
    float	    mult;	/* envelope scale factor (scale * volume) */
} VoiceEnvelope;
    
typedef struct _voiceLFO {
    MidiSine    value;
    MidiSine    step;
    float       scale;
} VoiceLFO;

#define MAX_PHASE   64

#define voiceDelay(v,i)    v->delay[ (i) & (MAX_PHASE-1) ]

typedef struct _midiVoice {
    /*
     * Assigned at NOTE_ON
     */
    unsigned char	channel;
    unsigned char	note;
    unsigned char	on;
    unsigned char	velocity;
    const Instrument	*instrument;
    const MidiPatch	*patch;
    const SampleElement	*se;

    /*
     * Computed while note plays
     */
    VoiceWave		wave;
    VoiceEnvelope	envelope;
    VoiceLFO		tremelo;
    VoiceLFO		vibrato;

    /*
     * phase shift
     */
    float	    delay[ MAX_PHASE ];
    int		    phase;
    int		    delay_pos;
} MidiVoice;


typedef struct _midiChannel {
    unsigned char   program;
    unsigned char   volume;
    unsigned char   expression;
    unsigned char   pan;
    unsigned char   sustain;
    unsigned char   drum;
    unsigned short  bend;
    unsigned char   bend_semi;
    unsigned char   bend_cent;
    unsigned char   modulation;
    unsigned char   chorus;
    int		    bend_range;
    double	    bend_ratio;
    int		    pn_registered;
    unsigned char   pn_r_msb;
    unsigned char   pn_r_lsb;
    unsigned char   pn_u_msb;
    unsigned char   pn_u_lsb;
} MidiChannel;


typedef struct _midi_output_Format {
    unsigned short  xor; /* for unsigned <-> signed conversion */
    int		    l_msb; /* buffer indices: */
    int		    l_lsb;
    int		    r_msb;
    int		    r_lsb;
    int		    stereo; /* true if output = stereo */
} Midi_output_format;

typedef struct _midiReverb {
    float	    left;
    float	    right;
} MidiReverb;

typedef struct _midiReverbTap {
    float	    amount;
    unsigned long   offset;
} MidiReverbTap;

typedef struct _midiData {
    /*
     * Midi source data 
     */
    void	    *source;
    struct _SourceFuncs	*sourceFuncs;
	int         repeat;

    /*
     * Audio output format
     */
    int		    dataRate;
    int		    stereo;

    int		    envelopeRate;
    int		    envelopeDiv;

    int		    dataBytesPerSample;
    int		    dataBytesPerUnit;	/* bytes per sample * channels */
    Midi_output_format format;

    int		    bufSamples;
    int		    maxVoices;

    /*
     * output queue
     */
    int		    put;
    int		    get;
    int		    queued;
    unsigned char   *buf;
    int		    syncQueue;	    /* no asynchronous queuing */

    /*
     * Midi state
     */
    int		    done;
    int		    midiDone;
    int		    nactive;
    int		    stopped;
    int		    paused;
    int		    divisions;
    int		    envelopeStep;
    int		    auto_pan;	    /* dynamic per-note pan */
    int		    chorus;
    float	    panPhase;
    double	    eventTime;	    /* tick position of next event */
    double	    time;	    /* current tick position */
    double	    midiSpeedFactor;
    double	    midiSpeed;	    /* ticks per sample */
    double	    lengthSeconds;  /* total song length in seconds */
    unsigned long   samples;	    /* current song position in samples */
    unsigned long   tempo;
    float	    scale;
    MidiVoice	    *voices;
    MidiChannel	    channels[ MIDI_CHANNELS ];
    MidiReverb	    *reverb;
    MidiReverbTap   *reverbTaps;
    char	    *reverbName;
    int		    reverbCount;
    unsigned long   reverbPos;
    unsigned long   reverbMask;

	double            noteTable[128]; /* WHS: moved here from a global in midi_data.c */

    /*
     * instrument data
     */
    const MidiMap   *map;
    const MidiPatch *patchMap [ 128 ];
    const MidiPatch *drumMap [ MAX_DRUM - MIN_DRUM + 1 ];
    char	    *patchFile;
} MidiData;

typedef enum _MidiState {
    midiStart,
    midiRunning,
    midiNoteOff,
    midiNoteOn,
    midiAfterTouch,
    midiControlChange,
    midiPitchWheel,
    midiSysex,
    midiSysexUniversalNonRealTime,
    midiSysexTuningNRT,
    midiSysexTuningBulkReplyMore,
    midiSysexTuningBulkReply,
    midiSysexUniversalRealTime,
    midiSysexDeviceControl,
    midiSysexTuningRT,
    midiSysexTuningRTNoteChange,
    midiSysexTuningRTNoteChangeMoreCont,
    midiSysexTuningRTNoteChangeMore,
    midiSysexFile,
    midiSysexFileMore,
    midiInt,
    midiIntMore,
    midiBufSome,
    midiSkipSome,
    midiSkip,
    midiEOF,
} MidiState;

typedef struct _MidiParseState {
    MidiState	    state;	    /* current parser state */
    MidiState	    returnState;    /* single element state stack */
    unsigned char   status;	    /* current running status */
    int		    value;	    /* general value */
    int		    count;	    /* counter, counts up to limit */
    int		    limit;
    Boolean	    obeyRealtime;   /* look for and parse System Realtime */
    unsigned char   buf[4];	    /* temporary storage */
} MidiParseState;


#if 1 /* might move to input/input.h */
typedef struct _SourceFuncs {
    int		    (*Read) (MidiData *d, char *);
    int		    (*Event) (MidiData *d, Boolean);
    void	    (*Dispose) (MidiData *d);
    void	    (*Reset) (MidiData *d);
    int		    (*Sample) (MidiData *d, char *);
} SourceFuncs;
#endif




/* Low level voice functions: */
void	voiceWaveInit (MidiData		*d,
		       MidiVoice	*v,
		       double		frequency,
		       Boolean		loopRelease,
		       float		pan);

#define voiceWaveChange(vw, freq)    ((vw)->sampleStep = (MidiFix) ((freq) * (vw)->freqFactor))

double	VoiceVolume (MidiData *d, int voice);
double	VoiceFrequency (MidiData *d, int voice);

Boolean	voiceEnvelopeInit (MidiData	    *d,
			   MidiVoice	    *v,
			   float	    volume,
			   float	    in_scale,
			   float	    pan);

void	voiceLFOInit (VoiceLFO	    *vl,
		      int32	    frequency,
		      unsigned long envelopeRate,
		      double	    scale);

void	voiceIdle (MidiData *d, int voice);

#define Volume(vel,vol,exp)    (((vel)*(vol)*(exp + 127)) / (127.0*127.0*254.0))

void	voicePress (MidiData	    *d,
		    int		    voice,
		    const MidiPatch *patch,
		    float	    volume,
		    float	    in_scale,
		    double	    frequency,
		    float	    pan);

void	voiceRelease (MidiData	    *d,
		      int	    voice,
		      int	    velocity);

void	voiceBend (MidiData	    *d,
		   int		    voice,
		   double	    frequency,
		   unsigned long    sampleRate);


void	voiceVolume (MidiData	    *d,
		     int	    voice);

void	voiceLFOChange (VoiceLFO	*vl,
		int32		frequency,
		unsigned long	envelopeRate,
		float		scale);

void	voiceLFOInit (VoiceLFO	    *vl,
	      int32	    frequency,
	      unsigned long envelopeRate,
	      double	    scale);


#define interpValue(s,f)	((s)[FixToInt(f)] * FixToFloor(f) + \
				 (s)[FixToInt(f)+1] * FixToCeil(f))
#define nearValue(s,f)		((s)[FixToInt(f)])

#define voiceWaveValueI(vw)	    interpValue((vw)->samples, (vw)->sample)
#define voiceWaveValueN(vw)	    nearValue((vw)->samples, (vw)->sample)

#define EnvelopeVolumeIndex(e)	((e) >> (31 - VOLUME_BITS))
#define EnvelopeVolume(e)  volumeTable[ EnvelopeVolumeIndex(e) ]

#define voiceEnvelopeValue(ve)	    (EnvelopeVolume((ve)->value) * \
				     (ve)->mult * (ve)->fTrem)


/* Higher level midi decoding/converting functions */
void	MidiOutInit (MidiData *d);

void    MidiParse (MidiData *d, char *s, int len);
int     MidiParseOne (MidiData *d, MidiParseState *, unsigned char *, int, Boolean);
int     MidiParseInt (unsigned char *buf, unsigned long *i);
void    MidiAllNotesOff (MidiData *d);

int     MidiSample (MidiData *d, char *);
void    MidiQueue (MidiData *d, int amount);

void    MidiSetFormat (MidiData *d, Boolean sixteenBit, 
					   Boolean unsign, Boolean msbFirst, 
					   Boolean stereo, Boolean leftFirst);

Boolean InitMidiPatch (MidiData *d);
void    FreeMidiPatch (MidiData *d);

int     MidiInit (MidiData *d);
void    MidiSetDefaults (MidiData *d);
double  MidiSeek (MidiData *d, double usec);
void    MidiReset (MidiData *d);

int     MidiSetFile (MidiData *d, char *);
void    MidiDispose (MidiData *d);
void    MidiFini (MidiData *d);
void    MidiClose (MidiData *d);
Boolean MidiSetReverbType (MidiData *d, char *name);

void    MidiSetTuning (MidiData *d, const double [12]);

Boolean MidiSetTemperment (MidiData *d, char *temperment);

void    MidiChangeTuning (MidiData *d, int note, int base, int cents);
void    MidiChangePanPhase (MidiData *d, float panPhase);
void    MidiSetNote (MidiData *d, int note, const double table[12], int base, int cents);

void
MidiNoteOn (MidiData	    *d,
			unsigned char   channel,
			unsigned char   note,
			unsigned char   velocity,
			Boolean	    seek);

void
MidiNoteOff (MidiData	    *d,
			 unsigned char  channel,
			 unsigned char  note,
			 unsigned char  velocity);


/* Low level midi/synth functions */
void MidiComputeBend (MidiData *d, unsigned char channel);
void MidiComputeTremelo (MidiData *d, unsigned char channel);
void MidiComputeVibrato (MidiData *d, unsigned char channel);
void MidiSetBendRange (MidiData *d, unsigned char channel);
void MidiChangeVolume (MidiData *d, unsigned char channel);
void MidiResetControllers (MidiData *d, int channel);

/* Low level midi parsing */
void MidiInitState (MidiParseState *s);

/* Internal: Instrument loading/writing */
const MidiPatch	*MidiPatchParse (const char *);
const MidiPatch	*ReadGUSPatch (const char *);

/* Instrument loading/writing */
const MidiPatch *MidiLoadDrum (MidiData *d, unsigned char note);
const MidiPatch *MidiLoadProgram (MidiData *d, unsigned char program);

int	PatchWrite (const char	    *patch,
				const MidiMap   *map,
				Boolean	    writeInstruments,
				Boolean	    writeBinary);


/* Per instrument volume: */
void	InitMidiVolumes (void);
void	SaveMidiVolumes (void);

#endif /* _GSI__gsimidi_pcm__kpmidi_h */
